#lang plai

(define-type FAE
  [num (n number?)]
  [add (lhs FAE?)
       (rhs FAE?)]
  [sub (lhs FAE?)
       (rhs FAE?)]
  [id (name symbol?)]
  [fun (param symbol?)
       (body FAE?)]
  [app (fun-expr FAE?)
       (arg-expr FAE?)]
  [if0 (test FAE?)
       (then FAE?)
       (else FAE?)])

#|
(define-type CFAE
8  [cnum (n number?)]
9  [cadd (lhs CFAE?)
        (rhs CFAE?)]
10  [csub (lhs CFAE?)
        (rhs CFAE?)]
11  [cid (pos number?)]
12  [cfun (body CFAE?)]
13  [capp (fun-expr CFAE?)
       (arg-expr CFAE?)]
14  [cif0 (test CFAE?)
        (then CFAE?)
        (else CFAE?)])
|#

#|
(define-type CFAE-Value
15  [numV (n number?)]
16  [closureV (body CFAE?)
            (ds DefrdSub?)])
|#

(define-type CDefrdSub
  [mtCSub]
  [aCSub (name symbol?)
         (rest CDefrdSub?)])

#|
(define-type CFAE-Cont
0  [mtK]
1  [addSecondK (r CFAE?)
              (ds DefrdSub?)
              (k CFAE-Cont?)]
2  [doAddK (v1 CFAE-Value?)
          (k CFAE-Cont?)]
3  [subSecondK (r CFAE?)
              (ds DefrdSub?)
              (k CFAE-Cont?)]
4  [doSubK (v1 CFAE-Value?)
          (k CFAE-Cont?)]
5  [appArgK (arg-expr CFAE?)
           (ds DefrdSub?)
           (k CFAE-Cont?)]
6  [doAppK (fun-val CFAE-Value?)
          (k CFAE-Cont?)]
7  [doIfK (then-expr CFAE?)
         (else-expr CFAE?)
         (ds DefrdSub?)
         (k CFAE-Cont?)])
|#

;; A DefrdSub is a list of CFAE-Value
; (define (DefrdSub? x) true)

;; ----------------------------------------

(define (any? x) true)

(define-type Pair
  [kons (first any?)
        (rest any?)])

(define fst kons-first)
(define rst kons-rest)

;; ----------------------------------------

;; parse : S-expr -> FAE
(define (parse sexp)
  (cond
    [(number? sexp) (num sexp)]
    [(symbol? sexp) (id sexp)]
    [(pair? sexp)
     (case (car sexp)
       [(+) (add (parse (second sexp)) (parse (third sexp)))]
       [(-) (sub (parse (second sexp)) (parse (third sexp)))]
       [(fun) (fun (first (second sexp)) (parse (third sexp)))]
       [(if0) (if0 (parse (second sexp))
                   (parse (third sexp))
                   (parse (fourth sexp)))]
       [else (app (parse (first sexp)) (parse (second sexp)))])]))

(test (parse 3) (num 3))
(test (parse 'x) (id 'x))
(test (parse '{+ 1 2}) (add (num 1) (num 2)))
(test (parse '{- 1 2}) (sub (num 1) (num 2)))
(test (parse '{fun {x} x}) (fun 'x (id 'x)))
(test (parse '{1 2}) (app (num 1) (num 2)))
(test (parse '{if0 0 1 2}) (if0 (num 0) (num 1) (num 2)))

;; ----------------------------------------

;; compile : FAE CDefrdSub -> CFae
(define (compile a-fae ds)
  (type-case FAE a-fae
    [num (n) (kons 8 n)]
    [add (l r) (kons 9 (kons (compile l ds) (compile r ds)))]
    [sub (l r) (kons 10 (kons (compile l ds) (compile r ds)))]
    [id (name) (kons 11 (locate name ds))]
    [fun (param body-expr) (kons 12
                                 (compile body-expr (aCSub param ds)))]
    [app (fun-expr arg-expr)
         (kons 13 (kons (compile fun-expr ds)
                       (compile arg-expr ds)))]
    [if0 (test-expr then-expr else-expr)
         (kons 14
               (kons (compile test-expr ds)
                     (kons
                      (compile then-expr ds)
                      (compile else-expr ds))))]))

(define (locate name ds)
  (type-case CDefrdSub ds
    [mtCSub () (error 'locate "free variable")]
    [aCSub (sub-name rest-sc)
           (if (symbol=? sub-name name)
               0
               (+ 1 (locate name rest-sc)))]))

;; ----------------------------------------

(define fae-reg (kons 8 0))
(define ds-reg 0)

;; interp : -> void
(define (interp)
  (case (fst fae-reg)
    [(8) ; num
     (begin
       (set! v-reg (kons 15 (rst fae-reg)))
       (continue))]
    [(9) ; add
     (begin
       (set! k-reg (kons 1
                         (kons (rst (rst fae-reg))
                               (kons ds-reg k-reg))))
       (set! fae-reg (fst (rst fae-reg)))
       (interp))]
    [(10) ; sub
     (begin
       (set! k-reg (kons 3
                         (kons (rst (rst fae-reg))
                               (kons ds-reg k-reg))))
       (set! fae-reg (fst (rst fae-reg)))
       (interp))]
    [(11) ; id
     (begin
       (set! ds2-reg ds-reg)
       (set! v-reg (rst fae-reg))
       (sc-ref))]
    [(12) ; fun
     (begin
       (set! v-reg (kons 16 (kons (rst fae-reg) ds-reg)))
       (continue))]
    [(13) ; app
     (begin
       (set! k-reg (kons 5
                         (kons (rst (rst fae-reg))
                               (kons ds-reg k-reg))))
       (set! fae-reg (fst (rst fae-reg)))
       (interp))]
    [(14) ; if0
     (begin
       (set! k-reg (kons 7
                         (kons (fst (rst (rst fae-reg)))
                               (kons (rst (rst (rst fae-reg)))
                                     (kons ds-reg k-reg)))))
       (set! fae-reg (fst (rst fae-reg)))
       (interp))]))

(define k-reg (kons 0 0))
(define v-reg 0)

;; continue : -> void
(define (continue)
  (case (fst k-reg)
    [(0) ; mtk
     v-reg]
    [(1) ; addSecondK
     (begin
       (set! fae-reg (fst (rst k-reg)))
       (set! ds-reg (fst (rst (rst k-reg))))
       (set! k-reg (kons 2 
                         (kons v-reg (rst (rst (rst k-reg))))))
       (interp))]
    [(2) ; doAddK
     (begin
       (set! v-reg (num+ (fst (rst k-reg)) v-reg))
       (set! k-reg (rst (rst k-reg)))
       (continue))]
    [(3) ; subSecondK
     (begin
       (set! fae-reg (fst (rst k-reg)))
       (set! ds-reg (fst (rst (rst k-reg))))
       (set! k-reg (kons 4 
                         (kons v-reg (rst (rst (rst k-reg))))))
       (interp))]
    [(4) ; doSubK
     (begin
       (set! v-reg (num- (fst (rst k-reg)) v-reg))
       (set! k-reg (rst (rst k-reg)))
       (continue))]
    [(5) ; appArgK
     (begin
       (set! fae-reg (fst (rst k-reg)))
       (set! ds-reg (fst (rst (rst k-reg))))
       (set! k-reg (kons 6 
                         (kons v-reg (rst (rst (rst k-reg))))))
       (interp))]
    [(6) ; doAppK
     (begin
       (set! fae-reg (fst (rst (fst (rst k-reg)))))
       (set! ds-reg (kons v-reg
                          (rst (rst (fst (rst k-reg))))))
       (set! k-reg (rst (rst k-reg)))
       (interp))]
    [(7) ; doIfK
     (begin
       (if (numzero? v-reg)
           (set! fae-reg (fst (rst k-reg)))
           (set! fae-reg (fst (rst (rst k-reg)))))
       (set! ds-reg (fst (rst (rst (rst k-reg)))))
       (set! k-reg (rst (rst (rst (rst k-reg)))))
       (interp))]))

(define ds2-reg 0)

(define (sc-ref)
  (if (zero? v-reg)
      (begin
        (set! v-reg (fst ds2-reg))
        (continue))
      (begin
        (set! ds2-reg (rst ds2-reg))
        (set! v-reg (- v-reg 1))
        (sc-ref))))

;; num-op : (number number -> number) -> (FAE-Value FAE-Value -> FAE-Value)
(define (num-op op op-name)
  (lambda (x y)
    (kons 15 (op (rst x) (rst y)))))

(define num+ (num-op + '+))
(define num- (num-op - '-))

;; numzero? : FAE-Value -> boolean
(define (numzero? x)
  (zero? (rst x)))

;; ----------------------------------------

(define init-k (kons 0 0))
(define (interpx a-fae ds k)
  (begin
    (set! fae-reg a-fae)
    (set! ds-reg ds)
    (set! k-reg k)
    (interp)))
(define (numV x) (kons 15 x))
(define empty-ds 0)

(test (interpx (compile (num 10) (mtCSub))
               empty-ds
               init-k)
      (numV 10))
(test (interpx (compile (add (num 10) (num 7)) (mtCSub))
              empty-ds
              init-k)
      (numV 17))
(test (interpx (compile (sub (num 10) (num 7)) (mtCSub))
              empty-ds
              init-k)
      (numV 3))
(test (interpx (compile
               (app (fun 'x (add (id 'x) (num 12)))
                    (add (num 1) (num 17)))
               (mtCSub))
              empty-ds
              init-k)
      (numV 30))
(test (interpx (compile (id 'x)
                       (aCSub 'x (mtCSub)))
              (kons (numV 10) empty-ds)
              init-k)
      (numV 10))

(test (interpx (compile (app (fun 'x (add (id 'x) (num 12)))
                            (add (num 1) (num 17)))
                       (mtCSub))
              empty-ds
              init-k)
      (numV 30))

(test (interpx (compile (app (fun 'x
                                 (app (fun 'f
                                           (add (app (id 'f) (num 1))
                                                (app (fun 'x
                                                          (app (id 'f) (num 2)))
                                                     (num 3))))
                                      (fun 'y (add (id 'x) (id 'y)))))
                            (num 0))
                       (mtCSub))
              empty-ds
               init-k)
      (numV 3))

(test (interpx (compile (if0 (num 0)
                             (num 1)
                             (num 2))
                        (mtCSub))
               empty-ds
               init-k)
      (numV 1))
(test (interpx (compile (if0 (num 1)
                             (num 0)
                             (num 2))
                        (mtCSub))
               empty-ds
               init-k)
      (numV 2))

(test (interpx (compile
                (parse 
                 '{{fun {mkrec}
                        {{fun {fib}
                              ;; Call fib on 4:
                              {fib 4}}
                         ;; Create recursive fib:
                         {mkrec
                          {fun {fib}
                               ;; Fib:
                               {fun {n}
                                    {if0 n
                                         1
                                         {if0 {- n 1}
                                              1
                                              {+ {fib {- n 1}}
                                                 {fib {- n 2}}}}}}}}}}
                   ;; mkrec:
                   {fun {body-proc}
                        {{fun {fX}
                              {fX fX}}
                         {fun {fX}
                              {body-proc {fun {x} {{fX fX} x}}}}}}})
                (mtCSub))
               empty-ds
               init-k)
      (numV 5))

(test/exn (compile (id 'x) (mtCSub))
          "free variable")
